/* Autor: Leandro Bernsmller

   mail: lebe@aton.inf.ufrgs.br

   visit my site: http://www.geocities.com/SiliconValley/Bay/1704

   Use this file as you want.

   Compile with Turbo C++ 3.0.

Camera com rotacao absoluta da base do sistema de referencia da camera
e translacao da origem (VOAR).
- multiplica a matriz transposta da base pela matriz de rotacao;
- o plano de projecao estah transladado para frente no SRU e o centro de
  projecao eh o centro da camera;
- permite projecoes atras do plano de projecoes;
- SRU mao direita e SRC mao esquerda.

             |Y                          Y
             |
             |                           |    Z
             |                          j|   /   SRC
             |                           |  /k
             |                           | /
             |                           |/
             |          X                +------> X
             +-----------                O      i
            /O                                    k=(0,0,-1)
           /            SRU
        Z /
         /
*/



#include <stdio.h>
#include <graphics.h>
#include <conio.h>
#include <math.h>



#define NVERTICES 8
#define DISTCP 500
#define CENTROX 320
#define CENTROY 240
#define RESX 640
#define RESY 480
#define INC 8



/* Exibe ou apaga o objeto
**********************************************************/
void Exibe ( float* pafX2D, float* pafY2D, int nCor )
{
   setcolor ( nCor );

   // faz um circulo entorno do primeiro vertice
   circle ( pafX2D [0], pafY2D [0], 5 );

   line ( pafX2D [0], pafY2D [0], pafX2D [1], pafY2D [1] );
   line ( pafX2D [0], pafY2D [0], pafX2D [3], pafY2D [3] );
   line ( pafX2D [0], pafY2D [0], pafX2D [7], pafY2D [7] );
   line ( pafX2D [1], pafY2D [1], pafX2D [2], pafY2D [2] );
   line ( pafX2D [1], pafY2D [1], pafX2D [6], pafY2D [6] );
   line ( pafX2D [2], pafY2D [2], pafX2D [3], pafY2D [3] );
   line ( pafX2D [2], pafY2D [2], pafX2D [5], pafY2D [5] );
   line ( pafX2D [3], pafY2D [3], pafX2D [4], pafY2D [4] );
   line ( pafX2D [4], pafY2D [4], pafX2D [5], pafY2D [5] );
   line ( pafX2D [4], pafY2D [4], pafX2D [7], pafY2D [7] );
   line ( pafX2D [5], pafY2D [5], pafX2D [6], pafY2D [6] );
   line ( pafX2D [6], pafY2D [6], pafX2D [7], pafY2D [7] );
   // a marca
   line ( pafX2D [0], pafY2D [0], pafX2D [2], pafY2D [2] );
   line ( pafX2D [1], pafY2D [1], pafX2D [3], pafY2D [3] );
}



/* Rotaciona pontos definidos no espaco 3D conforme
os 3 angulos (um para cada eixo).
**********************************************************/
void Rotaciona ( float fAngX, float fAngY, float fAngZ,
                 float& fX, float& fY, float& fZ )
{
   // numero PI
   float
      fPi = (float) 3.14159265359;

   // angulos em radianos
   float
      fRadX = fPi * fAngX / 180,
      fRadY = fPi * fAngY / 180,
      fRadZ = fPi * fAngZ / 180;

   float
      fXAux, fYAux, fZAux;

   fXAux = fX    * cos ( fRadZ ) + fY    * sin ( fRadZ );
   fYAux = fY    * cos ( fRadZ ) - fX    * sin ( fRadZ );
   fY    = fYAux * cos ( fRadX ) + fZ    * sin ( fRadX );
   fZAux = fZ    * cos ( fRadX ) - fYAux * sin ( fRadX );
   fX    = fXAux * cos ( fRadY ) + fZAux * sin ( fRadY );
   fZ    = fZAux * cos ( fRadY ) - fXAux * sin ( fRadY );
}



/* Transforma uma coordenada no espaco conforme uma base
de um sistema de referencia.
**********************************************************/
void AplicaBase ( float* pafI, float* pafJ, float* pafK,
                  float& fX, float& fY, float& fZ )
{
   float
      fXa = fX * pafI [ 0 ] + fY * pafI [ 1 ] + fZ * pafI [ 2 ],
      fYa = fX * pafJ [ 0 ] + fY * pafJ [ 1 ] + fZ * pafJ [ 2 ],
      fZa = fX * pafK [ 0 ] + fY * pafK [ 1 ] + fZ * pafK [ 2 ];
   fX = fXa;
   fY = fYa;
   fZ = fZa;
}



main ()
{
   // inicia modo grafico
   int
      nGrDriver = DETECT,
      nGrMode;
   initgraph ( &nGrDriver, &nGrMode, "" );

   // dados dos objetos nos sistemas de coordenadas
   float
      // objeto: cubo centrado na origem (SRU)
      afXU [ NVERTICES ] = { -50, 50,  50, -50, -50,  50,  50, -50 },
      afYU [ NVERTICES ] = {  50, 50,  50,  50, -50, -50, -50, -50 },
      afZU [ NVERTICES ] = {  50, 50, -50, -50, -50, -50,  50,  50 },
      // objeto no sistema de referencia de camera
      afXC [ NVERTICES ],
      afYC [ NVERTICES ],
      afZC [ NVERTICES ],
      // vertices projetados
      afX2D [ NVERTICES ] = { 0, 0, 0, 0, 0, 0, 0, 0 },
      afY2D [ NVERTICES ] = { 0, 0, 0, 0, 0, 0, 0, 0 },
      // base do sist. de ref. da camera: 3 vetores unitarios ortogonais
      afI [] = { 1, 0, 0 },     // vetor I com suas componentes: [ Xi Yi Zi ]
      afJ [] = { 0, 1, 0 },     // vetor J com suas componentes: [ Xj Yj Zj ]
      afK [] = { 0, 0, -1 },    // vetor K com suas componentes: [ Xk Yk Zk ]
      // origem do SRC relativo ao SRU
      afOCam [] = { 0, 0, 500 + DISTCP };

   // enquanto nao for teclado ESC exibe objeto
   int
      nFim = 0;
   while ( ! nFim ) {

      // exibe informacoes sobre camera
      gotoxy ( 1, 1 );
      printf ( "Use: setas cursor + teclado numerico: '2', '4', '6', '7', '8', '9' + 'a', 'z'.\n" );
      printf ( "Origem Camera:    x: %f   y: %f   z: %f\n", afOCam [ 0 ], afOCam [ 1 ], afOCam [ 2 ] );
      printf ( "I:(%f, %f, %f)\nJ:(%f, %f, %f)\nK:(%f, %f, %f) \n", afI [ 0 ], afI [ 1 ], afI [ 2 ],
          afJ [ 0 ], afJ [ 1 ], afJ [ 2 ], afK [ 0 ], afK [ 1 ], afK [ 2 ] );
      printf ( "Vertice 1 (SRU): x: %f   y: %f   z: %f\n", afXU [ 0 ], afYU [ 0 ], afZU [ 0 ] );
      printf ( "Vertice 1 (SRC): x: %f   y: %f   z: %f\n", afXC [ 0 ], afYC [ 0 ], afZC [ 0 ] );
      printf ( "Vertice 1 eh o circulado\n" );

      // trata teclado se alguma tecla foi pressionada
      if ( kbhit () ) {

         float
            // angulos de rotacao da base da camera a cada interacao
            fAngX = 0,
            fAngY = 0,
            fAngZ = 0;

         // le tecla
         char
            yTecla = getch ();

         // se foi teclado ESC
         if ( yTecla == 27 )
            nFim = 1;

         // controles de rotacao e translacao

         // se foi teclado para frente altera componente Z do vetor
         if ( yTecla == 'a' ) {
            // atualiza posicao da camera no SRU
            afOCam [ 0 ] += INC * afK [ 0 ];
            afOCam [ 1 ] += INC * afK [ 1 ];
            afOCam [ 2 ] += INC * afK [ 2 ];
         }

         // se foi teclado para tras altera componente Z do vetor
         if ( yTecla == 'z' ) {
            // atualiza posicao da camera no SRU
            afOCam [ 0 ] -= INC * afK [ 0 ];
            afOCam [ 1 ] -= INC * afK [ 1 ];
            afOCam [ 2 ] -= INC * afK [ 2 ];
         }

         // se foi teclado para direita altera componente X do vetor
         if ( yTecla == '6' ) {
            // atualiza posicao da camera no SRU
            afOCam [ 0 ] += INC * afI [ 0 ];
            afOCam [ 1 ] += INC * afI [ 1 ];
            afOCam [ 2 ] += INC * afI [ 2 ];
         }

         // se foi teclado para esquerda altera componente X do vetor
         if ( yTecla == '4' ) {
            // atualiza posicao da camera no SRU
            afOCam [ 0 ] -= INC * afI [ 0 ];
            afOCam [ 1 ] -= INC * afI [ 1 ];
            afOCam [ 2 ] -= INC * afI [ 2 ];
         }

         // se foi teclado para acima altera componente Y do vetor
         if ( yTecla == '8' ) {
            // atualiza posicao da camera no SRU
            afOCam [ 0 ] += INC * afJ [ 0 ];
            afOCam [ 1 ] += INC * afJ [ 1 ];
            afOCam [ 2 ] += INC * afJ [ 2 ];
         }

         // se foi teclado para baixo altera componente Y do vetor
         if ( yTecla == '2' ) {
            // atualiza posicao da camera no SRU
            afOCam [ 0 ] -= INC * afJ [ 0 ];
            afOCam [ 1 ] -= INC * afJ [ 1 ];
            afOCam [ 2 ] -= INC * afJ [ 2 ];
         }

         // se foi teclado para direita no teclado numerico
         if ( yTecla == '7' )
            fAngZ = 1;

         // se foi teclado para esquerda no teclado numerico
         if ( yTecla == '9' )
            fAngZ = -1;

         // se foi teclado setas (rotacao)
         if ( kbhit () ) {
            // para cima
            if ( ( yTecla = getch () ) == 72 )
               fAngX = -1;
            // para baixo
            if ( yTecla == 80 )
               fAngX = 1;
            // para direita
            if ( yTecla == 77 )
               fAngY = 1;
            // para esquerda
            if ( yTecla == 75 )
               fAngY = -1;
         }

         // Transforma base do sistema de referencia multiplicando
         // a matriz transposta da base pela matriz de rotacao.
         Rotaciona ( fAngX, fAngY, fAngZ, afI [ 0 ], afJ [ 0 ], afK [ 0 ] );
         Rotaciona ( fAngX, fAngY, fAngZ, afI [ 1 ], afJ [ 1 ], afK [ 1 ] );
         Rotaciona ( fAngX, fAngY, fAngZ, afI [ 2 ], afJ [ 2 ], afK [ 2 ] );

         // translada do objeto conforme origem do SRC em relacao ao SRU
         for ( int ncT = 0; ncT < NVERTICES; ncT ++ ) {

            afXC [ ncT ] = afXU [ ncT ] - afOCam [ 0 ];
            afYC [ ncT ] = afYU [ ncT ] - afOCam [ 1 ];
            afZC [ ncT ] = afZU [ ncT ] - afOCam [ 2 ] + DISTCP;
         }

         // aplica base ao objeto fazendo rotacao do SRU para SRC
         for ( ncT = 0; ncT < NVERTICES; ncT ++ )
            AplicaBase ( afI, afJ, afK, afXC [ ncT ], afYC [ ncT ], afZC [ ncT ] );

         // apaga objeto
         Exibe ( afX2D, afY2D, 0 );

         // projeta e ajusta para viewport
         for ( int ncP = 0; ncP < NVERTICES; ncP ++ ) {

            afZC [ ncP ] -= DISTCP;

            // projeta
            // verifica se vertice estah na frente do plano de projecao
            if ( afZC [ ncP ] > - DISTCP ) {
               afX2D [ ncP ] = afXC [ ncP ] / ( afZC [ ncP ] / DISTCP + 1 );
               afY2D [ ncP ] = afYC [ ncP ] / ( afZC [ ncP ] / DISTCP + 1 );
            }
            else {
               afX2D [ ncP ] = afXC [ ncP ] / ( afZC [ ncP ] / ( 1 - afZC [ ncP ] ) + 1 );
               afY2D [ ncP ] = afYC [ ncP ] / ( afZC [ ncP ] / ( 1 - afZC [ ncP ] ) + 1 );
            }

            // ajusta para viewport - meio da tela e inverte Y
            afX2D [ ncP ] = afX2D [ ncP ] + CENTROX;
            afY2D [ ncP ] = RESY - ( afY2D [ ncP ] + CENTROY );
         }

         // exibe objeto
         Exibe ( afX2D, afY2D, 15 );
      }
   }

   // termina modo grafico
   closegraph ();
}
